
/*
	Copyright (C) 2000, 2001  Charles MacDonald

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef _NES_ONLY


#include <windows.h>
#include <windowsx.h>
#include <shlwapi.h>

#include "arc.h"
#include "debug.h"
#include "crc32.h"


#include "pce.h"

#include "h6280.h"
#include "vdc.h"
#include "vce.h"
#include "render.h"
#include "system.h"
#include "psg.h"
#include "cpuintrf.h"
//#include "osd.h"
#include "h6280ops.h"

#include "nnnkaillera.h"	//nnn
extern int nnnKailleraFlag;
extern int nnnKailleraplayer;


#ifdef PCE_FAST_MEM
extern "C" {
uint8 *pce_ram;
uint8 *pce_read_ptr[8];	 /* Memory read pointers */
uint8 *pce_write_ptr[8];	/* Memory write pointers */
}
#endif


extern int MystrCmp(char *,char *);
extern void MyCtRestore();

static PCE *g_PCE;







PCE::PCE(const char* ROM_name, PCE_screen_mgr* _screen_mgr, sound_mgr* _sound_mgr, HWND parent_window_handle){
	scr_mgr = _screen_mgr;
	snd_mgr = _sound_mgr;
	window_handle = parent_window_handle;
	//	scr_mgr->setParentNES(this);
	g_PCE = this;
	apu_qreadp = apu_qwritep = 0;
	
	KailleraCounter = 0;
	y_offset=0;
	byr=0;
	vram_data_latch = 0;
	dvssr_trigger = 0;
	playfield_shift = 6;
	playfield_col_mask = 0xFF;
	playfield_row_mask = 0x1F;
	vramw = (uint16 *)&vram[0];
	objramw = (uint16 *)&objram[0];
	old_width = 0;
	old_height = 0;
	playfield_shift_table[0] = 6;
	playfield_shift_table[1] = 7;
	playfield_shift_table[2] = 8;
	playfield_shift_table[3] = 8;
	playfield_row_mask_table[0] = 0x1F;
	playfield_row_mask_table[1] = 0x3F;
	playfield_row_mask_table[2] = 0x7F;
	playfield_row_mask_table[3] = 0x7F;
	
	joy_sel = 0;
	joy_clr = 0;
	joy_cnt = 0;
	bitmap.depth = 16;

	/* Bit 0 : BG enable, Bit 1 : OBJ enable */
	plane_enable = -1;
//	render_line = NULL;

	
	try {
		if(!render_init())
			throw NULL;

		if(!LoadROM(window_handle, (char *)ROM_name))
			throw NULL;
	} catch(...) {
/*
		if(ROM!=NULL){
			free(ROM);
			ROM=NULL;
		}
*/
		throw;
	}
//	reset;
//	PCEp = this;
	reset(0);
}





//------------------------------



BOOL PCE::LoadROM(HWND hWnd, char *FileName)
{
	HANDLE hFile;
	DWORD dwBytes ,FileSize, i;
	unsigned char *pt = NULL, *p;
	const char *PCE_Extentions[] = { "*.pce", NULL};
	char *extp;
		
	extp=PathFindExtension(FileName);
	if(extp==NULL)
		return -1;
	++extp;
	for(i=0; PCE_Extentions[i]; i++){
		if(MystrCmp(extp, (char *)&PCE_Extentions[i][2])){
			break;
		}
	}
	if(PCE_Extentions[i]){
		hFile = CreateFile(FileName, GENERIC_READ, 0, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		if(hFile == INVALID_HANDLE_VALUE)
			return 0;
		FileSize = GetFileSize(hFile, NULL);
		pt = (unsigned char *)malloc(FileSize);
		if(pt==NULL){
			CloseHandle(hFile);
			return 0;
		}
		ReadFile(hFile, pt, FileSize, &dwBytes, NULL);
		CloseHandle(hFile);
	}
	else{
		if(!Uncompress(NULL, &pt, FileName, (LPDWORD)&FileSize, PCE_Extentions))
			return 0;
	}
	
	RomCRC = CrcCalc(pt, FileSize);

//	ROM = pt;
//	ROM_size = FileSize/0x2000;
	int flip, split;
	flip=	0;
	split=	0;


	p = pt+(FileSize%0x2000);
#if 1
	/* Always split 384K images */
	if((FileSize & ~0x1ffff) == 0x60000)
	{
		memcpy(rom + 0x00000, p + 0x00000, 0x40000);
		memcpy(rom + 0x80000, p + 0x40000, 0x20000);
	}
	else /* Split 512K images if requested */
	if(split && ((FileSize & ~0x1ffff) == 0x80000))
	{
		memcpy(rom + 0x00000, p + 0x00000, 0x40000);
		memcpy(rom + 0x80000, p + 0x40000, 0x40000);
	}
	else
	{
		memcpy(rom, p, (FileSize > 0x100000) ? 0x100000 : FileSize-512);
	}
#else
	memcpy(rom, p, FileSize);
#endif
	free(pt);


	ROMTitle[0] = 0;
	ROM_size = FileSize;

	strcpy(ROMPath, FileName);
	PathRemoveFileSpec(ROMPath);
	{
		char *fnp = PathFindFileName(FileName);
		if(fnp!=NULL){
			strcpy(ROMname, fnp);
			PathRemoveExtension(ROMname);
		}
		else{
			strcpy(ROMname, "n/a");
		}
	}
	return TRUE;
}



PCE::~PCE(){
/*
	if(ROM!=NULL){
		free(ROM);
		ROM=NULL;
	}
*/
	pce_shutdown();
	vdc_shutdown();
	psg_shutdown();
	render_shutdown();
	return;
}


void PCE::new_snd_mgr(sound_mgr* _sound_mgr){
	snd_mgr = _sound_mgr;
	audio_init(snd_mgr->get_sample_rate());
}



void PCE::reset(unsigned char f)
{
//	system_init(snd_mgr->get_sample_rate());
//	system_init(0);
	audio_init(snd_mgr->get_sample_rate());
	pce_reset();
	vdc_reset();
	psg_reset();
	render_reset();
}


void PCE::GetROMInfoStr(char *wt){
	wsprintf(wt, "filesize = %u, CRC32 = %08X", ROM_size, RomCRC);
}


void PCE::Save_SaveRAM(){

}


void PCE::Load_SaveRAM(){

}
/*
static int framelen = 16743;
static int framecount;
*/


boolean PCE::emulate_frame(boolean draw)
{
	int lockf = 0;
//---	vid_begin();
	if(nnnKailleraFlag)
		KailleraClientProc();
	MyCtRestore();

	pixmap p;
#if 1
	if(draw && (lockf = scr_mgr->lock(p))){
		bitmap.data = p.data;
		bitmap.pitch = p.pitch;
		bitmap.width = p.width;
		bitmap.height = p.height;
		bitmap.viewport.x =0;
		bitmap.viewport.y =0;
		bitmap.granularity=2;
	}
#endif	
	int line;
	h6280_set_execute_cycles(0);

	for(y_offset = byr, line = 0; line < 262; line += 1)
	{
		if((line + 64) == (reg[6] & 0x3FF))
		{
			if(reg[5] & 0x04)
			{
				status |= STATUS_RR;
				h6280_set_irq_line(0, ASSERT_LINE);
			}
		}

		/* VBlank */
		if(line == 240)
		{
			if(dvssr_trigger || (reg[0x0F] & 0x10))
			{
				/* Clear DVSSR write trigger */
				dvssr_trigger = 0;

				/* Copy VRAM to object RAM */
				memcpy(objram, &vram[(reg[0x13] << 1) & 0xFFFE], 0x200);

				/* Cause transfer complete interrupt if necessary */
				if(reg[0x0F] & 0x01)
				{
					status |= STATUS_DS;
					h6280_set_irq_line(0, ASSERT_LINE);
				}

				/* Precalculate sprite data for the next frame */
				make_sprite_list();
			}

			/* Cause VBlank interrupt if necessary */
			if(reg[5] & 0x0008)
			{
				status |= STATUS_VD;
				h6280_set_irq_line(0, ASSERT_LINE);
			}
		}

		/* 7.16 MHz = 455 cycles per line */
		h6280_execute(455); 

		/* Render a line of the display */

		if((line < bitmap.height) && (lockf)){
			render_line(line);
		}

		/* Update internal line counter and wrap */
		y_offset = (y_offset + 1) & playfield_col_mask;

	}
	if(lockf){
		scr_mgr->unlock();
		lockf = 0;
	}

	/* Update audio */
//	if(snd.enabled) psg_update(snd.buffer[0], snd.buffer[1], snd.buffer_size);
	apu_process();
	return 0;
}



void PCE::apu_enqueue(pce_apudata_t *d)
{
/*
	if(apu_qwritep==apu_qreadp){
#ifdef _DEBUG
		LOG("apu_enqueue error" <<endl);
#endif
//		return;
	}
*/
	apu_queue[apu_qwritep] = *d;
	apu_qwritep= (apu_qwritep+1)&PCE_APUQUEUE_MASK;
}


pce_apudata_t *PCE::apu_dequeue(void)
{
#ifdef _DEBUG
//	LOG("apu_qreadp " << HEX(apu_qreadp,8) <<"  apu_qwritep " << HEX(apu_qwritep,8) <<endl);
#endif
//	if(apu_qreadp == apu_qwritep)
//		return NULL;
	pce_apudata_t *t;
	t = &apu_queue[apu_qreadp];
	apu_qreadp= (apu_qreadp+1)&PCE_APUQUEUE_MASK;
	return t;
}


static int mseq(DWORD *rand_val){
	if(*rand_val & 0x00080000){
		*rand_val = ((*rand_val ^ 0x0004) << 1) + 1;
		return 1;
	}
	else{
		*rand_val <<= 1;
		return 0;
	}
}

static long ftofix(double x){ 
	if (x > 32767.0) {
		return 0x7FFFFFFF;
	}
	if (x < -32767.0) {
		return -0x7FFFFFFF;
	}
	
	return (long)(x * 65536.0 + (x < 0 ? -0.5 : 0.5)); 
}

void PCE::apu_process(){
	unsigned char *buffer;
	int num_samples;
	int elapsed_cycles, inc_cycles, buffpos=0;
	int i, s_rate;
	pce_apudata_t *d;
	uint32 buf_len;

	elapsed_cycles = 0;

	s_rate = snd_mgr->get_sample_rate();
	if(snd_mgr->IsNull())
		return;
	{
		sound_buf_pos cur_ph;  // cur playing half
		sound_buf_pos cur_nph; // cur not-playing half

		cur_ph = snd_mgr->get_currently_playing_half();
		if(cur_ph == currently_playing_half) return;
		cur_nph = currently_playing_half;
		currently_playing_half = cur_ph;
		if(!snd_mgr->lock(cur_nph, (void**)&buffer, &buf_len)){
//			apu_qreadp = apu_qwritep =0;
//			LOG("couldn't lock sound buffer" << endl);
			return;
		}
#ifdef _DEBUG
//		LOG("lock sound buffer " << HEX(buf_len,8) <<endl);
//		gb_sndclock = 0;
#endif
		num_samples = buf_len/(snd_mgr->get_sample_bits()/8*2);
//		memcpy(buf, pcm.buf, buf_len);
	}
//	for (; gb_cpu.snd >= RATE; gb_cpu.snd -= RATE)

//	inc_cycles = (1<<21/*119210*/) / s_rate; //4194304 / num_samples;//<<1;
	inc_cycles = h6280_get_execute_cycles()*2 / num_samples; //4194304 / num_samples;//<<1;
#ifdef _DEBUG
//	LOG("all " << HEX(h6280_get_execute_cycles(),8) << ", inc " << HEX(inc_cycles,8) << endl);
#endif
#if 0
	for(i=0;i<num_samples;i++, elapsed_cycles+=inc_cycles){
		int l, r;
		l=r= 0;

		int ch;					/* Channel index */

		while((apu_qreadp != apu_qwritep) /*&& (apu_queue[apu_qreadp].timestamp <= elapsed_cycles)*/){

			d = apu_dequeue();
			psg_w(d->address, d->value);
		}

		int start;				 /* Skip channels 0, 1 if LFO is enabled */ 
		int stop;				  /* Skip channels 4, 5 if noise is enabled */

		start = ((psg.lfoctrl & 3) == 0) ? 0 : 2;
		stop = 6; //(psg.noisectrl & 0x80) ? 4 : 6;

		for(ch = start; ch < stop; ch += 1)
		{
			/* If channel is ON and DDA is OFF, play waveform data */
			if((psg.channel[ch].control & 0xC0) == 0x80 && !(ch>=4 && !(psg.channel[ch].noisectrl & 0x80)))
			{
				/* Global sound balance (left and right, all channels) */
				int lbal = (psg.globalbalance >> 4) & 0x0F;
				int rbal = (psg.globalbalance >> 0) & 0x0F;

				/* Balance (left and right, this channel) */
				int lchb = (psg.channel[ch].balance >> 4) & 0x0F;
				int rchb = (psg.channel[ch].balance >> 0) & 0x0F;

				/* Volume level (this channel) */
				int chvl = (psg.channel[ch].control & 0x1F);

				/* Total volume levels for left and right
				   (volume sounds too soft - not sure how to combine these) */
				int lvol = (lbal + lchb + chvl);
				int rvol = (rbal + rchb + chvl);

				int base, step, offset, data;

				/* This is the largest possible step value which is divided
				   by the channel frequency used to increment the counter,
				   which in turn is used to traverse the waveform buffer.
				   3580000 (PSG clock) / 32 (length of waveform) = 111875 (base step value)
				   That doesn't work right but multiplying it by three sounds better. */
				base = (3580000 / 32) * 3;

				/* Calculate the value to add to the counter for each sample,
				   but don't divide by zero if the frequency is zero */
				step = (psg.channel[ch].frequency) ? base / psg.channel[ch].frequency : 0;

				/* Use upper 5 bits of 12-bit frequency as wave index */
				offset = (psg.channel[ch].counter >> 12) & 0x1F;

				/* Bump waveform index */
				psg.channel[ch].counter += step;

				/* Data is 5 bits */
				data = (psg.channel[ch].waveform[offset] & 0x1F);

				/* Add new sample to old one */
				l = (l + (lvol * data * 2));
				r = (r + (rvol * data * 2));
			}
		}
		/* Make samples signed */
		if(l & 0x8000) l ^= 0x8000;
		if(r & 0x8000) r ^= 0x8000;

		/* Store samples in buffer */
		*buffer++ = (l>>8) + 128;
		*buffer++ = (r>>8) + 128;
//		*buffer++ = l+128;
//		*buffer++ = r+128;

	}
#else
	{
		const DWORD rand_val[6] = {0,0,0,0,0x51F631E4,0x51F631E4};
		static DWORD	ka[6] = {0,0,0,0,0,0};
		static DWORD	na[6] = {0,0,0,0,0,0};
		static DWORD	ra[6] = {0,0,0,0,0,0};
		static int t, lf=0, rf=0;
		static int wave[32];

	/* a*2^(b*x) */
	/* a = 400, b = log2(16384/a), x = 0.0 ~ 1.0 */
		const int vol_tbl[32] =
		{
			100,451,508,573,646,728,821,925,
			1043,1175,1325,1493,1683,1898,2139,2411,
			2718,3064,3454,3893,4388,4947,5576,6285,
			7085,7986,9002,10148,11439,12894,14535,16384,
		};

		int j;
		int l, r;
		int ch;					/* Channel index */

		for(i=0;i<num_samples;i++, elapsed_cycles+=inc_cycles){
			l=r= 0;
			while((apu_qreadp != apu_qwritep)&& (apu_queue[apu_qreadp].timestamp <= elapsed_cycles)){
#ifdef _DEBUG
//				LOG("c " << HEX(d->timestamp,8)  << endl);
#endif
				d = apu_dequeue();
				psg_w(d->address, d->value);
			}

			for(ch = 0; ch < 6; ch += 1)
			{
				if (!(psg.channel[ch].control&0x80))
					continue;
				int lvol, rvol;
				lvol = rvol = 0;


				if(psg.channel[ch].control&0x40){
					wave[0] =  (psg.channel[ch].waveform[0]/*-16*/)*512;//*702;
					lvol = ((psg.globalbalance>>3)&0x1E) + (psg.channel[ch].control&0x1F) + ((psg.channel[ch].balance>>3)&0x1E);
					lvol = lvol-60;
					if(lvol < 0)
						lvol = 0;
					lvol = wave[0]*vol_tbl[lvol]/16384;
					rvol = ((psg.globalbalance<<1)&0x1E) + (psg.channel[ch].control&0x1F) + ((psg.channel[ch].balance<<1)&0x1E);
					rvol = rvol-60;
					if(rvol < 0)
						rvol = 0;
					rvol = wave[0]*vol_tbl[rvol]/16384;

					l+=lvol;
					r+=rvol;
				}
				else
				{
//					continue;

					//for (dwPos = 0; dwPos < dwSize; dwPos++)
					//	*buf++ = 0;
					//return;
					if (ch >= 4 && (psg.channel[ch].noisectrl&0x80)){	//Noise
						DWORD Np = (psg.channel[ch].noisectrl&0x1F);
						lvol = ((psg.globalbalance>>3)&0x1E) + (psg.channel[ch].control&0x1F) + ((psg.channel[ch].balance>>3)&0x1E);
						lvol = lvol-60;
						if (lvol < 0)
							lvol = 0;
						lvol = vol_tbl[lvol];
						rvol = ((psg.globalbalance<<1)&0x1E) + (psg.channel[ch].control&0x1F) + ((psg.channel[ch].balance<<1)&0x1E);
						rvol = rvol-60;
						if (rvol < 0)
							rvol = 0;
						rvol = vol_tbl[rvol];

						ka[ch] += 3000+Np*512;
						t = ka[ch] / (DWORD)s_rate;
						if (t >= 1){
							ra[ch] = mseq((DWORD *)&rand_val[ch]);
							ka[ch] -= s_rate*t;
						}
						l+= ((ra[ch] ? 10* 512/*702*/ : -10*512 /*702*/)*lvol/16384);
						r+= ((ra[ch] ? 10* 512/*702*/ : -10*512 /*702*/)*rvol/16384);
					}
					else{

						for(j = 0; j < 32; j++)
							wave[j] = (psg.channel[ch].waveform[j]/*-16*/ )* /*702*/ 512;

						int Tp = psg.channel[ch].frequency;
						if(Tp != 0){
							lvol = ((psg.globalbalance>>3)&0x1E) + (psg.channel[ch].control&0x1F) + ((psg.channel[ch].balance>>3)&0x1E);
							lvol = lvol-60;
							if (lvol < 0)
								lvol = 0;
							lvol = vol_tbl[lvol];
							rvol = ((psg.globalbalance<<1)&0x1E) + (psg.channel[ch].control&0x1F) + ((psg.channel[ch].balance<<1)&0x1E);
							rvol = rvol-60;
							if (rvol < 0)
								rvol = 0;
							rvol = vol_tbl[rvol];

							l+= (wave[na[ch]]*lvol/16384);
							r+= (wave[na[ch]]*rvol/16384);

							ka[ch] += /*N*/32*1118608/Tp;
							t = ka[ch] / (10*(DWORD)s_rate);
							na[ch] = (na[ch]+t)%32/*N*/;
							ka[ch] -= 10*s_rate*t;
						}
					}
				}
			}


#if 0
			{
				int tr, tl;
				tr = r;
				tl = l;

				l = (l+lf)>>1;
				r = (r+rf)>>1;

				lf = tl;
				rf = tr;
			}
#endif
			if (l > 0x7FFF)
				l = 0x7FFF;
			else if (l < -0x8000)
				l = -0x8000;
			if (r > 0x7FFF)
				r = 0x7FFF;
			else if (l < -0x8000)
				r = -0x8000;
			*buffer++ = (l >> 8) ^ 0x80;
			*buffer++ = (r >> 8) ^ 0x80;
/*
			char tl = (l>>8);
			if(tl<-128) tl=-128;
			if(tl>127) tl=127;
			tl+=0x80;
			*buffer++ = tl;
			char tr = (r>>8);
			if (tr<-128) tr=-128;
			if (tr>127) tr=127;
			tr+=0x80;
			*buffer++ = tr;
*/
		}
/*
		p = buffer;
		pw = (WORD *)psoundbuf;
		for(j=0; j<num_samples; j++){
			char tl = (*pw>>8);
			pw++;
			if(tl<-128) tl=-128;
			if(tl>127) tl=127;
			tl+=0x80;
			*p++ = tl;
			char tr = (*pw>>8);
			pw++;
			if (tr<-128) tr=-128;
			if (tr>127) tr=127;
			tr+=0x80;
			*p++ = tr;
//			if(*pw & 0x8000) *pw ^= 0x8000;
//			*p++ = (*pw>>8) + 128;
//			pw++;
//			if(*pw & 0x8000) *pw ^= 0x8000;
//			*p++ = (*pw>>8) + 128;
//			pw++;
		}
*/

		/* Make samples signed */
/*
		if(l & 0x8000) l ^= 0x8000;
		if(r & 0x8000) r ^= 0x8000;
*/
		/* Store samples in buffer */
//		*buffer++ = (l>>8) + 128;
//		*buffer++ = (r>>8) + 128;
//		*buffer++ = l+128;
//		*buffer++ = r+128;

	}
#endif

	snd_mgr->unlock();
#ifdef _DEBUG
//	LOG("__ apu_qreadp " << HEX(apu_qreadp,8) <<"  apu_qwritep " << HEX(apu_qwritep,8) <<endl);
#endif
	{
		int tmp;
#if 0
		for(tmp = apu_qreadp; tmp != apu_qwritep; tmp= (tmp+1)&PCE_APUQUEUE_MASK){
			apu_queue[tmp].timestamp -= elapsed_cycles/*-inc_cycles*/;
		}
#else
		for(tmp = apu_qreadp; tmp != apu_qwritep; tmp= (tmp+1)&PCE_APUQUEUE_MASK){
			apu_queue[tmp].timestamp = 0;
		}
#endif
	}
}


////



int pce_irq_callback(int irqline)
{
#ifdef _DEBUG
//	LOG("pce_irq_callback  " << HEX(irqline,8) <<endl);
#endif
	return (0);
}

/*--------------------------------------------------------------------------*/
/* Init, reset, shutdown functions										  */
/*--------------------------------------------------------------------------*/


void PCE::pce_reset(void)
{
	joy_sel = joy_clr = joy_cnt = 0;
	memset(ram, 0, 0x8000);
	memset(cdram, 0, 0x10000);
	memset(&vce, 0, sizeof(vce));
	memset(&CDROM2, 0, sizeof(CDROM2));
#ifdef PCE_FAST_MEM
	memset(pce_dummy, 0, 0x2000);
	bank_reset();
#endif
//	load_file("pce.brm", bram, 0x2000);
	h6280_reset(0);
	h6280_set_irq_callback(&pce_irq_callback);
}

void PCE::pce_shutdown(void)
{
//	if(save_bram) save_file("pce.brm", bram, 0x2000);
#ifdef DEBUG
	error("PC:%04X\n", h6280_get_pc());
#endif	
}




/*--------------------------------------------------------------------------*/
/* Memory handlers														  */
/*--------------------------------------------------------------------------*/

#ifndef PCE_FAST_MEM

void PCE::cpu_writeport16(int port, int data)
{
	vdc_w(port, data);
}

void PCE::cpu_writemem21(int address, int data)
{
	uint8 page = (address >> 13) & 0xFF;

	/* RAM (F8) */
	if(page == 0xF8 || page == 0xF9 || page == 0xFA || page == 0xFB) {
		ram[(address & 0x7FFF)] = data;
		return;
	}

	/* I/O (FF) */
	if(page == 0xFF) {
		io_page_w(address & 0x1FFF, data);
		return;
	}

	/* CD RAM */
	if((page >= 0x80) && (page <= 0x87)) {
		cdram[(address & 0xFFFF)] = data;
		return;
	}

	/* Backup RAM (F7) */
	if(page == 0xF7) {
		bram[(address & 0x1FFF)] = data;
		return;
	}

#ifdef DEBUG
	error("write %02X to %02X:%04X (%08X)\n", data, page, address & 0x1fff, h6280_get_reg(H6280_PC));
#endif
}

int PCE::cpu_readmem21(int address)
{
	uint8 page;

	if(address <= 0x0FFFFF) return (rom[(address)]);

	page = (address >> 13) & 0xFF;

	/* ROM (00-7F) */
	if(page <= 0x7F) return (rom[(address)]);

	/* RAM (F8) */
	if(page == 0xF8 || page == 0xF9 || page == 0xFA || page == 0xFB) return (ram[(address & 0x7FFF)]);

	/* I/O (FF) */
	if(page == 0xFF) return (io_page_r(address & 0x1FFF));

	/* CD RAM */
	if((page >= 0x80) && (page <= 0x87)) return (cdram[(address & 0xFFFF)]);

	/* Backup RAM (F7) */
	if(page == 0xF7) return (bram[(address & 0x1FFF)]);

#ifdef DEBUG
	error("read %02X:%04X (%08X)\n", page, address & 0x1fff, h6280_get_reg(H6280_PC));
#endif
	return (0xFF);
}

#endif


/*--------------------------------------------------------------------------*/
/* Hardware page handlers												   */
/*--------------------------------------------------------------------------*/

void PCE::io_page_w(int address, int data)
{
	switch(address & 0x1C00)
	{
		case 0x0000: /* VDC */
//			if(address <= 0x0003)
			{
				vdc_w(address&3, data);
				return;
			}
			break;

		case 0x0400: /* VCE */
//			if(address <= 0x0405)
			{
				vce_w(address&0x407, data);
				return;
			}
			break;

		case 0x0800: /* PSG */
//			if(address <= 0x0809)
#ifdef _DEBUG
//			LOG("A " << HEX(address, 4) << "D " << HEX(data, 2) << endl);
#endif
			{
#if 0
				psg_w(address, data); 
#else
				pce_apudata_t	t;
				t.address	= address&0x80F;
				t.timestamp	= h6280_get_execute_cycles();
				t.value		= data;
				apu_enqueue(&t);
#endif
				return;
			};
			break;

		case 0x0C00: /* Timer */
//			if(address == 0x0C00 || address == 0x0C01)
#ifdef _DEBUG
//			LOG("A " << HEX(address, 4) << "D " << HEX(data, 2) << endl);
#endif
			{
				H6280_timer_w(address & 1, data);
				return;
			};
			break;

		case 0x1000: /* I/O */
//			if(address == 0x1000)
			{
				input_w(data);
				return;
			}
			break;

		case 0x1400: /* IRQ control */
//			if(address == 0x1402 || address == 0x1403)
			{
				int a = address & 3;
				if(a&2)
					H6280_irq_status_w(address & 1, data);
				return;
			};
			break;

		case 0x1800: /* CD-ROM */
			CDROM_IO_W(address, data);
#ifdef DEBUG
			error("cdrom %04X = %02X\n", address, data);
#endif
//			if(address&0x0f == 3/*0x1807*/) save_bram = 1;
			break;

		case 0x1C00: /* Expansion */
			break;
	}

#ifdef DEBUG
	error("write %02X to IO %04X (%08X)\n", data, address, h6280_get_reg(H6280_PC));
#endif
}


int PCE::io_page_r(int address)
{
	switch(address & 0x1C00)
	{
		case 0x0000: /* VDC */
//			if(address <= 0x0003)
				return (vdc_r(address&3));
			break;

		case 0x0400: /* VCE */
//			if(address <= 0x0405)
				return (vce_r(address&0x407));
			break;

		case 0x0800: /* PSG */
#ifdef _DEBUG
	//		LOG("Read IO 0x0800" /*<< HEX(apu_qreadp,8) <<"  apu_qwritep " << HEX(apu_qwritep,8) */<<endl);
#endif
			break;

		case 0x0C00: /* Timer */
//			if(address == 0x0C00 || address == 0x0C01)
				return (H6280_timer_r(address & 1));
			break;

		case 0x1000: /* I/O */
//			if(address == 0x1000)
				return (input_r());
			break;

		case 0x1400: /* IRQ control */
			{
				int a = address & 3;
				if(a&2)
					return (H6280_irq_status_r(address & 1));
			}
//			if(address == 0x1402 || address == 0x1403)
//				return (H6280_irq_status_r(address & 1));
			break;

		case 0x1800: /* CD-ROM */
			return CDROM_IO_R(address);
#ifdef DEBUG
			error("cdrom %04X\n", address);
#endif
			return 0xff;
			break;

		case 0x1C00: /* Expansion */
			break;
	}

#ifdef DEBUG
	error("read IO %04X (%08X)\n", address, h6280_get_reg(H6280_PC));
#endif
	return (0x00);
}

/*--------------------------------------------------------------------------*/
/* Input routines														   */
/*--------------------------------------------------------------------------*/

void PCE::input_w(uint8 data)
{
	joy_clr = (data >> 1) & 1;
	if(joy_clr){
		joy_cnt=0;
		joy_sel=1;
	}
	else{
		if((data & 1) && !joy_sel){
			joy_cnt++;
			joy_cnt%=5;
		}
		joy_sel = (data & 1);
	}
}

uint8 PCE::input_r(void)
{
	uint8 temp = *Pad[joy_cnt];

/*
	input.pad[joy_cnt] = *Pad;
	if(input.pad[joy_cnt] & INPUT_LEFT)   temp &= ~0x80;
	if(input.pad[joy_cnt] & INPUT_DOWN)   temp &= ~0x40;
	if(input.pad[joy_cnt] & INPUT_RIGHT)  temp &= ~0x20;
	if(input.pad[joy_cnt] & INPUT_UP)	 temp &= ~0x10;
	if(input.pad[joy_cnt] & INPUT_RUN)	temp &= ~0x08;
	if(input.pad[joy_cnt] & INPUT_SELECT) temp &= ~0x04;
	if(input.pad[joy_cnt] & INPUT_B2)	 temp &= ~0x02;
	if(input.pad[joy_cnt] & INPUT_B1)	 temp &= ~0x01;
*/
	if(joy_sel & 1)
		temp >>= 4;
	temp &= 0x0F;

	/* Set D6 for TurboGrafx-16, clear for PC-Engine */
	if(input.system & SYSTEM_TGX)
		temp |= 0x40;

	return (temp);
}




uint8 PCE::CDROM_IO_R(int address){
#ifdef _DEBUG
//	LOG("CDROM_IO_R  " << HEX(address, 4) <<"  ,"  <<endl);
#endif
#if 1
	switch(address & 0x0f){
		case 0x00:
			return CDROM2.CD_IO[0];
			break;
		case 0x01:
			break;
		case 0x02:
			return CDROM2.CD_IO[2];
			break;
		case 0x03:
			save_bram = 0;
			break;
		case 0x04:
			return CDROM2.CD_IO[4];
			break;
		case 0x05:
			break;
		case 0x06:
			break;
		case 0x07:
			break;
		case 0x08:
			break;
		case 0x09:
			break;
		case 0x0A:
			break;
		case 0x0B:
			break;
		case 0x0C:
			break;
		case 0x0D:
			break;
		case 0x0E:
			break;
		case 0x0F:
			break;
	}
#endif
	return 0xff;
}



void PCE::CDROM_IO_W(int address, uint8 data){
#ifdef _DEBUG
	LOG("CDROM_IO_W  " << HEX(address, 4) <<"  ," << HEX(data,2) <<endl);
#endif
	switch(address & 0x0f){
		case 0x00:
			if(data == 0x81)
				CDROM2.CD_IO[0] = 0xD0;
//			CDROM2.CD_1800H = data;
			break;
		case 0x01:
			CDROM2.CD_IO[0] = data;
			switch(data){
				case 0x00:
					break;
				case 0x03:
					break;
				case 0x08:
					break;
				case 0x81:
					CDROM2.CD_IO[0] = 0x40;
					break;
				case 0xD8:
				case 0xD9:
				case 0xDA:
					
				case 0xDD:
				case 0xDE:
					break;
			}
			break;
		case 0x02:
			if(!(CDROM2.CD_IO[2]&0x80) && (data & 0x80)){
				CDROM2.CD_IO[0] &= ~0x40;
			}
			else if((CDROM2.CD_IO[2]&0x80) && !(data & 0x80)){
				CDROM2.CD_IO[0] |= 0x40;
			}

			CDROM2.CD_IO[2] = data;
			break;
		case 0x03:
			break;
		case 0x04:
			if(data & 0x02){	// Reset
				data &= 0x0fd;
				WRMEM(0x222D, 1);
			}
			else{
				CDROM2.CD_IO[4] = data;
			}
			break;
		case 0x05:
			break;
		case 0x06:
			break;
		case 0x07:
//			if(data & 0x80)
				save_bram = 1;
			break;
		case 0x08:
			break;
		case 0x09:
			break;
		case 0x0A:
			break;
		case 0x0B:
			break;
		case 0x0C:
			break;
		case 0x0D:
			break;
		case 0x0E:
			break;
		case 0x0F:
			break;
	}
}



#ifdef PCE_FAST_MEM

void PCE::bank_reset(void)
{
	int i;
	pce_ram = ram;
	for(i = 0; i < 8; i += 1)
	{
		pce_read_ptr[i] = &rom[0x0000];
		pce_write_ptr[i] = &pce_dummy[0x0000];
	}
}

void PCE::bank_set(int bank, int value)
{
	/* ROM */
	if(value <= 0x7F) {
		pce_read_ptr[bank] = &rom[(value << 13)];
		pce_write_ptr[bank] = &pce_dummy[0x0000];
		return;
	}

	/* CD RAM */
	if((value >= 0x80) && (value <= 0x87)) {
		pce_read_ptr[bank] = pce_write_ptr[bank] = &cdram[(value & 0x07) << 13];
		return;
	}

	/* RAM */
	if((value >= 0xF8) && (value <= 0xFB)) {
		pce_read_ptr[bank] = pce_write_ptr[bank] = &ram[(value & 0x03) << 13];
		return;
	}

	/* Backup RAM */
	if(value == 0xF7) {
		pce_read_ptr[bank] = pce_write_ptr[bank] = &bram[0x0000];
		return;
	}

	/* I/O page */
	if(value == 0xFF) {
		pce_read_ptr[bank] = pce_write_ptr[bank] = NULL;
		return;
	}

	/* Unknown page */
#ifdef DEBUG
	error("Map unknown page %02X to MMR #%d\n", value, bank);
#endif
	pce_read_ptr[bank] = pce_write_ptr[bank] = &pce_dummy[0x0000];
}

#endif







void PCE::audio_init(int rate)
{
	memset(&snd, 0, sizeof(snd));

	/* Exit if no sound or invalid sample rate */
	if(!rate || ((rate < 8000) || (rate > 44100))) return;

	snd.buffer_size = snd_mgr->get_buffer_len();
	snd.sample_rate = rate;

	/* Set audio enable flag */
	snd.enabled = 1;
}







void PCE::vce_w(int address, int data)
{
	int msb = (address & 1);

#ifdef _DEBUG
//	LOG("vce_w  " << HEX(address, 4) <<"  ," << HEX(data,2) <<endl);
#endif
	switch(address & ~1)
	{
		case 0x0404: /* Data */
			{
#if 0
				if(data != vce.data[((vce.addr & 0x1FF) << 1) | (msb)])
				{
					vce.data[((vce.addr & 0x1FF) << 1) | (msb)] = data&0x1FF;

					if((vce.addr & 0x0F) != 0x00)
					{
						uint16 temp = *(uint16 *)&vce.data[(vce.addr << 1)];

						pixel_16bit[(vce.addr & 0x1FF)] = pixel_lut_16bit[temp];
						pixel_32bit[(vce.addr & 0x1FF)] = pixel_lut_32bit[temp];
					}

					/* Update overscan color */
					if((vce.addr & 0x0F) == 0x00)
					{
						int n;
						uint16 temp = *(uint16 *)&vce.data[0];

						for(n = 0; n < 0x10; n += 1){
							pixel_16bit[(n << 4)] = pixel_lut_16bit[temp];
							pixel_32bit[(n << 4)] = pixel_lut_32bit[temp];
						}
/*
						temp = (temp >> 1) & 0xFF;
						for(n = 0; n < 0x10; n += 1)
							xlat[0][(n << 4)] = temp;
*/
					}
				}
#else
//				if(data != vce.data[((vce.addr & 0x1FF) << 1) | (msb)])
				{
					if(!msb){
						vce.data[((vce.addr & 0x1FF) << 1) ] = data&0x0FF;
					}
					else{
						vce.data[((vce.addr & 0x1FF) << 1) | 1] = data&0x01;
					}

					if((vce.addr & 0x0F) != 0x00)
					{
						uint16 temp = *(uint16 *)&vce.data[(vce.addr << 1)];

						pixel_16bit[(vce.addr & 0x1FF)] = pixel_lut_16bit[temp];
						pixel_32bit[(vce.addr & 0x1FF)] = pixel_lut_32bit[temp];
					}

					/* Update overscan color */
					if((vce.addr & 0x0F) == 0x00)
					{
						int n;
						uint16 temp = *(uint16 *)&vce.data[0];

						for(n = 0; n < 0x10; n += 1){
							pixel_16bit[(n << 4)] = pixel_lut_16bit[temp];
							pixel_32bit[(n << 4)] = pixel_lut_32bit[temp];
						}
					}
				}
#endif
			}

			/* Increment VCE address on access to the MSB data port */
			if(msb) vce.addr += 1;
			break;

		case 0x0402: /* Address */
			if(msb)
				vce.addr = (vce.addr & 0x00FF) | ((data & 1) << 8);
			else
				vce.addr = (vce.addr & 0x0100) | (data);
			break;

		case 0x0400: /* Control */
			if(!msb) vce.ctrl = (data & 1);
			break;
	}
}


int PCE::vce_r(int address)
{
	int msb = (address & 1);

	if((address & ~1) == 0x0404)
	{
		uint8 temp = vce.data[((vce.addr & 0x1FF) << 1) | (msb)];
		if(msb) vce.addr += 1;
		return (temp);
	}

	return (0xFF);
}



/*--------------------------------------------------------------------------*/
/* Init, reset, shutdown routines										   */
/*--------------------------------------------------------------------------*/

int PCE::psg_init(void)
{
	memset(&psg, 0, sizeof(psg));
	return (0);
}

void PCE::psg_reset(void)
{
	memset(&psg, 0, sizeof(psg));
	int i;
//	for(i=0; i<6; i++)
//		psg.channel[i].control = 0x80;
}

void PCE::psg_shutdown(void)
{
}

/*--------------------------------------------------------------------------*/
/* PSG emulation															*/
/*--------------------------------------------------------------------------*/

void PCE::psg_w(uint16 address, uint8 data)
{

#if 1
#ifdef _DEBUG
//	LOG("psg_w  " << HEX(address, 4) <<"  ," << HEX(data,2) <<endl);
#endif
	switch(address)
	{
		case 0x0800: /* Channel select */
			psg.select = (data & 7);
			break;

		case 0x0801: /* Global sound balance */
			psg.globalbalance = data;
			break;

		case 0x0802: /* Channel frequency (LSB) */
			PSGCH.frequency = (PSGCH.frequency & 0x0F00) | (data);
			break;

		case 0x0803: /* Channel frequency (MSB) */
			PSGCH.frequency = (PSGCH.frequency & 0x00FF) | ((data & 0x0F) << 8);
			break;

		case 0x0804: /* Channel enable, DDA, volume */
			PSGCH.control = data;
			if((data & 0xC0) == 0x40) PSGCH.waveform_index = 0;
//			if(data & 0x40) PSGCH.waveform_index = 0;
			break;

		case 0x0805: /* Channel balance */
			PSGCH.balance = data;
			break;

		case 0x0806: /* Channel waveform data */
			if(PSGCH.control & 0x40){
//				PSGCH.waveform_index=0;
//				PSGCH.waveform[PSGCH.waveform_index] = data;
				PSGCH.waveform[0] = data&0x1F;
			}
			else{
				PSGCH.waveform[PSGCH.waveform_index] = data&0x1F;
				PSGCH.waveform_index = ((PSGCH.waveform_index + 1) & 0x1F);
			}
			break;

		case 0x0807: /* Noise enable and frequency */
			PSGCH.noisectrl = data;
			break;

		case 0x0808: /* LFO frequency */
			psg.lfofreq = data;
			break;

		case 0x0809: /* LFO trigger and control */
			psg.lfoctrl = data;
			break;
	}
#endif
}
#if 0
void PCE::psg_update(int16 *bufl, int16 *bufr, int length)
{
	/* Fill as many samples as needed */
	while(length > 0)
	{
		int ch;					/* Channel index */
		int sample[2] = {0, 0};	/* Left and right samples */
		int start;				 /* Skip channels 0, 1 if LFO is enabled */ 
		int stop;				  /* Skip channels 4, 5 if noise is enabled */

		start = ((psg.lfoctrl & 3) == 0) ? 0 : 2;
		stop = (psg.noisectrl & 0x80) ? 4 : 6;

		for(ch = start; ch < stop; ch += 1)
		{
			/* If channel is ON and DDA is OFF, play waveform data */
			if((psg.channel[ch].control & 0xC0) == 0x80)
			{
				/* Global sound balance (left and right, all channels) */
				int lbal = (psg.globalbalance >> 4) & 0x0F;
				int rbal = (psg.globalbalance >> 0) & 0x0F;

				/* Balance (left and right, this channel) */
				int lchb = (psg.channel[ch].balance >> 4) & 0x0F;
				int rchb = (psg.channel[ch].balance >> 0) & 0x0F;

				/* Volume level (this channel) */
				int chvl = (psg.channel[ch].control & 0x1F);
	
				/* Total volume levels for left and right
				   (volume sounds too soft - not sure how to combine these) */
				int lvol = (lbal + lchb + chvl);
				int rvol = (rbal + rchb + chvl);

				int base, step, offset, data;

				/* This is the largest possible step value which is divided
				   by the channel frequency used to increment the counter,
				   which in turn is used to traverse the waveform buffer.
				   3580000 (PSG clock) / 32 (length of waveform) = 111875 (base step value)
				   That doesn't work right but multiplying it by three sounds better. */
				base = (3580000 / 32) * 3;

				/* Calculate the value to add to the counter for each sample,
				   but don't divide by zero if the frequency is zero */
				step = (psg.channel[ch].frequency) ? base / psg.channel[ch].frequency : 0;

				/* Use upper 5 bits of 12-bit frequency as wave index */
				offset = (psg.channel[ch].counter >> 12) & 0x1F;

				/* Bump waveform index */
				psg.channel[ch].counter += step;

				/* Data is 5 bits */
				data = (psg.channel[ch].waveform[offset] & 0x1F);

				/* Add new sample to old one */
				sample[0] = (sample[0] + (lvol * data));
				sample[1] = (sample[1] + (rvol * data));
			}
		}

		/* Make samples signed */
		if(sample[0] & 0x8000) sample[0] ^= 0x8000;
		if(sample[1] & 0x8000) sample[1] ^= 0x8000;

		/* Store samples in buffer */
		*bufl++ = sample[0];
		*bufr++ = sample[1];

		/* Do next sample pair */
		--length;
	}
}
#endif



#define LOG_DMA	 0

#define MARK_BG_DIRTY(addr)									 \
{															   \
	int name = (addr >> 4) & 0x7FF;							 \
	if(bg_name_dirty[name] == 0)								\
	{														   \
		bg_name_list[bg_list_index] = name;					 \
		bg_list_index += 1;									 \
	}														   \
	bg_name_dirty[name] |= (1 << (addr & 0x07));								 \
}

#define MARK_OBJ_DIRTY(addr)									\
{															   \
	int name = (addr >> 6) & 0x1FF;							 \
	if(obj_name_dirty[name] == 0)							   \
	{														   \
		obj_name_list[obj_list_index] = name;				   \
		obj_list_index += 1;									\
	}														   \
	obj_name_dirty[name] |= (1 << (addr & 0x0F));			   \
}



/*--------------------------------------------------------------------------*/
/* Memory handlers														  */
/*--------------------------------------------------------------------------*/

int PCE::vdc_r(int offset)
{
	uint8 temp;
	uint8 msb = (offset & 1);

	switch(offset)
	{
		case 0x0000: /* Register latch / status flags */
			temp = status;
			status = 0;
			h6280_set_irq_line(0, CLEAR_LINE);
			return (temp);

		case 0x0002: /* Data port (LSB) */
		case 0x0003: /* Data port (MSB) */
			if(latch == 0x02)
			{
				temp = (vram[((reg[1] << 1) | (msb)) & 0xFFFF]);
				if(msb) reg[1] += addr_inc;
				return (temp);
			}
			break;
	}

	return (0xFF);
}

void PCE::vdc_w(int offset, int data)
{
	uint8 msb = (offset & 1);

	switch(offset)
	{
		case 0x0000: /* Register latch / status flags */
			latch = (data & 0x1F);
			break;

		case 0x0002: /* Data port (LSB) */
		case 0x0003: /* Data port (MSB) */

			if(msb)
				reg[latch] = (reg[latch] & 0x00FF) | (data << 8);
			else
				reg[latch] = (reg[latch] & 0xFF00) | (data);

			switch(latch)
			{
				case 0x02: 
					if(msb)
					{
						/* Form complete VRAM word */
						uint16 vram_word = (data << 8 | vram_data_latch);

						/* Check if data is new or not */
						if(vram_word != vramw[(reg[0] & 0x7FFF)])
						{
							/* Write data to VRAM */
							vramw[(reg[0] & 0x7FFF)] = vram_word;

							/* Mark pattern dirty tables */
							MARK_BG_DIRTY(reg[0]);
							MARK_OBJ_DIRTY(reg[0]);
						}

						reg[0] += addr_inc;
					}
					else
					{
						vram_data_latch = data;
					}
					break;

				case 0x08:
					y_offset = byr = (reg[0x08] & 0x1FF);
					y_offset &= playfield_col_mask;
					break;

				case 0x05:
					if(msb) {
						static uint8 add_tbl[] = {1, 32, 64, 128};
						addr_inc = add_tbl[(data >> 3) & 3];
					}
					break;

				case 0x09:
					if(!msb) {
						playfield_shift = playfield_shift_table[(data >> 4) & 3];
						playfield_row_mask = playfield_row_mask_table[(data >> 4) & 3];
						playfield_col_mask = ((data >> 6) & 1) ? 0x01FF : 0x00FF;
					}
					break;

				case 0x0B:
					disp_width = (1+(reg[0x0B] & 0x3F)) << 3;
					disp_nt_width = (disp_width >> 3);

					if(disp_width != old_width) {
						bitmap.viewport.ow = bitmap.viewport.w;
						bitmap.viewport.w = old_width = disp_width;
						bitmap.viewport.changed = 1;
					}

					break;

				case 0x0D:
					disp_height = 1+(reg[0x0D] & 0x01FF);

					if(disp_height != old_height) {
						bitmap.viewport.oh = bitmap.viewport.h;
						bitmap.viewport.h = old_height = disp_height;
						bitmap.viewport.changed = 1;
					}
					break;

				case 0x12:
					if(msb) vdc_do_dma();
					break;

				case 0x13:
					if(msb) dvssr_trigger = 1;
					break;
			}
	}
}

/*--------------------------------------------------------------------------*/
/* Init, reset, shutdown routines										   */
/*--------------------------------------------------------------------------*/

int PCE::vdc_init(void)
{
	return (0);
}

void PCE::vdc_reset(void)
{
	memset(vram, 0, 0x10000);
	memset(reg, 0, 0x20);
	status = latch = 0;
	addr_inc = 1;
	dvssr_trigger = 0;

	playfield_shift = 6;
	playfield_row_mask = 0x1f;
	playfield_col_mask = 0xff;

	memset(bg_name_dirty, 0, sizeof(bg_name_dirty));
	memset(bg_name_list, 0, sizeof(bg_name_list));
	memset(bg_pattern_cache, 0, sizeof(bg_pattern_cache));
	bg_list_index = 0;

	memset(obj_name_dirty, 0, sizeof(obj_name_dirty));
	memset(obj_name_list, 0, sizeof(obj_name_list));
	memset(obj_pattern_cache, 0, sizeof(obj_pattern_cache));
	obj_list_index = 0;
}


void PCE::vdc_shutdown(void)
{
}


void PCE::vdc_do_dma(void)
{
	int did = (reg[0x0F] >> 3) & 1;
	int sid = (reg[0x0F] >> 2) & 1;
	int dvc = (reg[0x0F] >> 1) & 1;
	int sour = (reg[0x10] & 0x7FFF);
	int desr = (reg[0x11] & 0x7FFF);
	int lenr = (reg[0x12] & 0x7FFF);
/*
	int sour = reg[0x10];
	int desr = reg[0x11];
	int lenr = reg[0x12];
*/
#if LOG_DMA
	error("DMA S:%04X%c D:%04X%c L:%04X\n", sour, (sid) ? '-' : '+', desr, (did) ? '-' : '+', lenr);
#endif

	/* Do VRAM -> VRAM transfer and update pattern caches */
	do {
		uint16 temp = vramw[(sour & 0x7FFF)];

		if(temp != vramw[(desr & 0x7FFF)])
		{
			vramw[(desr & 0x7FFF)] = temp;
			MARK_BG_DIRTY(desr);
			MARK_OBJ_DIRTY(desr);
		}

		sour = (sid) ? (sour - 1) : (sour + 1);
		desr = (did) ? (desr - 1) : (desr + 1);
	} while (lenr--);

	/* Set VRAM -> VRAM transfer completed flag */
	status |= STATUS_DV;

	/* Cause IRQ1 if enabled */
	if(dvc)
	{
		h6280_set_irq_line(0, ASSERT_LINE);
	}
}


/*--------------------------------------------------------------------------*/



/*--------------------------------------------------------------------------*/
/* Init, reset, shutdown functions										  */
/*--------------------------------------------------------------------------*/


int PCE::render_init(void)
{
	int i, j, x;
	int lockf;
	pixmap p;

	lockf = scr_mgr->lock(p);
	if(lockf==0)
		return 0;
	scr_mgr->unlock();
	bitmap.data = p.data;
	bitmap.pitch = p.pitch;
	bitmap.width = p.width;
	bitmap.height = p.height;
	bitmap.viewport.x =0;
	bitmap.viewport.y =0;
	bitmap.granularity=2;

	bitdepth = p.bitdepth;
	if(p.bitdepth == 8)
		return 0;

	/* Make bitplane to pixel lookup table */
	for(i = 0; i < 0x100; i += 1)
		for(j = 0; j < 0x100; j += 1)
		{
			uint32 out = 0;
			for(x = 0; x < 8; x += 1)
			{
				out |= (j & (0x80 >> x)) ? (uint32)(8 << (x << 2)) : 0;
				out |= (i & (0x80 >> x)) ? (uint32)(4 << (x << 2)) : 0;
			}
#ifndef BIG_ENDIAN
			bp_lut[(j << 8) | (i)] = out;
#else
			bp_lut[(i << 8) | (j)] = out;
#endif
		}

	/* Set up pixel color remap tables, the alignment is DOS specific */
//	ptr = (short *)malloc(0x400);
//	if(!ptr) return (0);
/*	
	memset(rendtptr, 0, 0x400);
	xlat[0] = (uint8 *)(((uint32)rendtptr + 0x200) & ~0x1FF);
	xlat[1] = xlat[0] + 0x100;
*/
	/* Make VCE data to raw pixel look-up table */
	if(p.Gbitmask == 0x7E0){
		for(i = 0; i < 0x200; i += 1)
		{
			int r = (i >> 3) & 7;
			int g = (i >> 6) & 7;
			int b = (i >> 0) & 7;
			//		pixel_lut[i] = (r << 13 | g << 8 | b << 2) & 0xE71C;
			pixel_lut_16bit[i] = (r << 13 | g << 8 | b << 2);	//RRRRR GGGGGG BBBBB
		}
	}
	else{
		for(i = 0; i < 0x200; i += 1)
		{
			int r = (i >> 3) & 7;
			int g = (i >> 6) & 7;
			int b = (i >> 0) & 7;
			pixel_lut_16bit[i] = (r << 12 | g << 7 | b << 2);	//xRRRRR GGGGG BBBBB
//			pixel_lut_16bit[i] = (r << 13 | g << 8 | b << 2);	//RRRRR GGGGGG BBBBB
		}
	}

	for(i = 0; i < 0x200; i += 1)
	{
		int r = (i >> 3) & 7;
		int g = (i >> 6) & 7;
		int b = (i >> 0) & 7;
		pixel_lut_32bit[i] = (r << 21 | g << 13 | b << 5);	//RRRR RRRR GGGG GGGG BBBB BBBB
	}
//			render_func = pce_render_32bit;

	/*
	switch(p.bitdepth){
		case 16:
			for(i = 0; i < 0x200; i += 1)
			{
				int r = (i >> 3) & 7;
				int g = (i >> 6) & 7;
				int b = (i >> 0) & 7;
				//		pixel_lut[i] = (r << 13 | g << 8 | b << 2) & 0xE71C;
				//		pixel_lut[i] = (r << 12 | g << 7 | b << 2);	//xRRRRR GGGGG BBBBB
				pixel_lut_16bit[i] = (r << 13 | g << 8 | b << 2);	//RRRRR GGGGGG BBBBB
			}
//			render_func = pce_render_16bit;
			break;
		case 24:
		case 32:
			for(i = 0; i < 0x200; i += 1)
			{
				int r = (i >> 3) & 7;
				int g = (i >> 6) & 7;
				int b = (i >> 0) & 7;
				pixel_lut_32bit[i] = (r << 32 | g << 13 | b << 5);	//RRRR RRRR GGGG GGGG BBBB BBBB
			}
//			render_func = pce_render_32bit;
			break;
		default:
			return 0;
	}
*/
	return (1);
}


void PCE::render_reset(void)
{
	/* Hack for Mac port */
 //   render_line = (bitmap.depth == 8) ? render_line_8 : render_line_16;
}


void PCE::render_shutdown(void)
{
	/* todo: free xlat pointer */
}


int PCE::make_sprite_list(void)
{
	uint16 *sat = &objramw[0];
	int xpos, ypos, name, attr;
	int cgx, xflip, cgy, yflip;
	int width, height;
	int i;
	uint32 flip;

	used_sprite_index = 0;
	memset(&used_sprite_list, 0, sizeof(used_sprite_list));

	memset(&sprite_list, 0, sizeof(sprite_list));

	for(i = 0; i < 0x40; i += 1)
	{
		ypos = sat[(i << 2) | (0)];
		xpos = sat[(i << 2) | (1)];
		name = sat[(i << 2) | (2)];
		attr = sat[(i << 2) | (3)];

		ypos &= 0x3FF;
		xpos &= 0x3FF;

		if(xpos && ypos)
		{
			ypos -= 64;
			if(ypos >= 0x100) continue;
			cgy = (attr >> 12) & 3;
			cgy |= (cgy >> 1);
			height = (cgy + 1) << 4;
			if((ypos + height) < 0) continue;

			xpos -= 32;
			if(xpos >= 0x200) continue;
			cgx = (attr >> 8) & 1;
			width  = (cgx) ? 32 : 16;
			if((xpos + width) < 0) continue;

			xflip = (attr >> 11) & 1;
			yflip = (attr >> 15) & 1;
			flip = ((xflip << 9) | (yflip << 10)) & 0x600;

			name = (name >> 1) & 0x1FF;
			name &= ~((cgy << 1) | cgx);
			name |= flip;
			if(xflip && cgx) name ^= 1;

			sprite_list[i].top = ypos;
			sprite_list[i].bottom = ypos + height;
			sprite_list[i].xpos = xpos;
			sprite_list[i].name_left = name;
			sprite_list[i].name_right = name ^ 1;
			sprite_list[i].height = (height - 1);
			sprite_list[i].palette = (attr & 0x0F) << 4;

			if(yflip)
				sprite_list[i].flags |= FLAG_YFLIP;

			if(cgx)
				sprite_list[i].flags |= FLAG_CGX;

			if(!(attr & 0x80))
				sprite_list[i].flags |= FLAG_PRIORITY;

			used_sprite_list[used_sprite_index] = (i);
			used_sprite_index += 1;
		}
	}
/*
	{
		RECT zr;
		zr == 

		for(j = (used_sprite_index - 1); j >=0; j -= 1){
			t_sprite *p;
			int j = used_sprite_list[j];
			p = &sprite_list[i];
			p->
			if( (line >= p->top) && (line < p->bottom)){
				v_line = (line - p->top) & p->height;
				nt_line = v_line;
				if(p->flags & FLAG_YFLIP) nt_line = (p->height - nt_line);
				name_mask = ((nt_line >> 4) & 3) << 1;
				name = (p->name_left | name_mask);
				v_line &= 0x0F;
				if(p->flags & FLAG_CGX){
				}
			}
		}
	}
*/
	return (used_sprite_index);
}


/*--------------------------------------------------------------------------*/
/* Pattern and object cache update routines								 */
/*--------------------------------------------------------------------------*/

void PCE::update_bg_pattern_cache(void)
{
	int i;
	uint8 x, y, c;
	uint16 name, index1, index2;
	uint32 temp;

	if(!bg_list_index) return;

	for(i = 0; i < bg_list_index; i += 1)
	{
		name = bg_name_list[i];
		bg_name_list[i] = 0;

		for(y = 0; y < 8; y += 1)
		{
			if(bg_name_dirty[name] & (1 << y))
			{
				index1 = vramw[(name << 4) | (y)];
				index2 = vramw[(name << 4) | (y) | (8)];

				temp = (bp_lut[index1] >> 2) | bp_lut[index2];

				for(x = 0; x < 8; x += 1)
				{
					c = (temp >> (x << 2)) & 0x0F;
					bg_pattern_cache[(name << 6) | (y << 3) | (x)] = (c);
				}
			}
		}
		bg_name_dirty[name] = 0;
	}

	bg_list_index = 0;
}



void PCE::update_obj_pattern_cache(void)
{
	int i;
	uint16 name;
	uint16 b0, b1, b2, b3;
	uint8 i0, i1, i2, i3;		   
	uint8 x, y, c;

	if(!obj_list_index) return;

	for(i = 0; i < obj_list_index; i += 1)
	{
		name = obj_name_list[i];
		obj_name_list[i] = 0;

		for(y = 0; y < 0x10; y += 1)
		{
			if(obj_name_dirty[name] & (1 << y))
			{
				b0 = vramw[(name << 6) + (y) + (0x00)];
				b1 = vramw[(name << 6) + (y) + (0x10)];
				b2 = vramw[(name << 6) + (y) + (0x20)];
				b3 = vramw[(name << 6) + (y) + (0x30)];

				for(x = 0; x < 0x10; x += 1)
				{
					i0 = (b0 >> (x ^ 0x0F)) & 1;
					i1 = (b1 >> (x ^ 0x0F)) & 1;
					i2 = (b2 >> (x ^ 0x0F)) & 1;
					i3 = (b3 >> (x ^ 0x0F)) & 1;

					c = (i3 << 3 | i2 << 2 | i1 << 1 | i0);

					obj_pattern_cache[(name << 8) | (y << 4) | (x)] = (c);
					obj_pattern_cache[0x20000 | (name << 8) | (y << 4) | (x ^ 0x0F)] = (c);
					obj_pattern_cache[0x40000 | (name << 8) | ((y ^ 0x0F) << 4) | (x)] = (c);
					obj_pattern_cache[0x60000 | (name << 8) | ((y ^ 0x0F) << 4) | (x ^ 0x0F)] = (c);
				}
			}
		}
		obj_name_dirty[name] = 0;
	}
	obj_list_index = 0;
}


/*--------------------------------------------------------------------------*/
/* Render functions														 */
/*--------------------------------------------------------------------------*/


void PCE::render_line(int line)
{
	{
		for(int i =0; i < PCE_SCREEN_WIDTH_MAX; i++)
			Line_Cache[i] = 0;

	}
//	memset(Line_Cache, 0, sizeof(Line_Cache));

	if((reg[0x05] & 0x80) && (plane_enable & 1))
	{
		update_bg_pattern_cache();
		render_bg(line);
	}
/*
	else
	{
		int i;
		uint16 *ptr = (uint16 *)&bitmap.data[(line * bitmap.pitch) + (bitmap.viewport.x * bitmap.granularity)];
		for(i = 0; i < disp_width; i += 1)
			ptr[i] = pixel[0][0];			
	}
*/

	if((reg[0x05] & 0x40) && (plane_enable & 2))
	{
		update_obj_pattern_cache();
		render_obj(line);
	}

//	if(render_func)
//		render_func(line);
	if(bitdepth==16){
		// render 16 bit color
		uint16 *dst, *src;
		int i, x;
		
		src = Line_Cache + 0x20/* + shift*/;
		dst = (uint16 *)&bitmap.data[(line * bitmap.pitch)]; //((/*0x20 +*/ (0 - shift)) << 1)];
		
		x=disp_width;
		if(bitmap.width < x)
			x = bitmap.width;
		for(i = 0; i < x; i++){
			dst[i] = pixel_16bit[src[i] & 0x1ff];
		}
	}
	else if(bitdepth == 24){
		// render 32 bit color
		uint8 *dst;
		uint16 *src;
		int i, x;

		src = Line_Cache + 0x20/* + shift*/;
		dst = (uint8 *)&bitmap.data[(line * bitmap.pitch)]; //((/*0x20 +*/ (0 - shift)) << 1)];

		x=disp_width;
		if(bitmap.width < x)
			x = bitmap.width;
		for(i = 0; i < x; i++){
			dst[i*3] = *((uint8 *)(&pixel_32bit[src[i] & 0x1ff]));
			dst[i*3+1] = *((uint8 *)(&pixel_32bit[src[i] & 0x1ff])+1);
			dst[i*3+2] = *((uint8 *)(&pixel_32bit[src[i] & 0x1ff])+2);
		}
	}
	else {
		// render 32 bit color
		uint32 *dst;
		uint16 *src;
		int i, x;

		src = Line_Cache + 0x20/* + shift*/;
		dst = (uint32 *)&bitmap.data[(line * bitmap.pitch)]; //((/*0x20 +*/ (0 - shift)) << 1)];
		
		x=disp_width;
		if(bitmap.width < x)
			x = bitmap.width;
		for(i = 0; i < x; i++){
			dst[i] = pixel_32bit[src[i] & 0x1ff];
		}
	}
}


void PCE::render_bg(int line)
{
	uint16 *nt;
	uint8 *src, palette;
	uint16 *dst;
	int column, name, attr, x, shift, v_line, nt_scroll;
	int xscroll = (reg[7] & 0x03FF);
	int end = disp_nt_width;

	/* Offset in pattern, in lines */
	v_line = (y_offset & 7);

	/* Offset in name table, in columns */
	nt_scroll = (xscroll >> 3);

	/* Offset in column, in pixels */
	shift = (xscroll & 7);

	/* Draw an extra tile for the last column */
	if(shift) end += 1;

	/* Point to current offset within name table */
	nt = (uint16 *)&vram[(y_offset >> 3) << playfield_shift];

	/* Point to start in line buffer */
//	dst = (uint16 *)&bitmap.data[(line * bitmap.pitch) + ((/*0x20 +*/ (0 - shift)) << 1)];
	dst = Line_Cache + (0x20 - shift);

	/* Draw columns */
	for(column = 0; column < end; column += 1)
	{
		/* Get attribute */
		attr = nt[(column + nt_scroll) & playfield_row_mask];
		
		/* Extract name and palette bits */
		name = (attr & 0x07FF);
		palette = (attr >> 8) & 0xF0;
		
		/* Point to current pattern line */
		src = &bg_pattern_cache[(name << 6) + (v_line << 3)];
		
		/* Draw column */
		for(x = 0; x < 8; x += 1)
		{
			dst[(column << 3) | (x)] = (src[x] | palette);
		}
	}
}


void PCE::render_obj(int line)
{
	t_sprite *p;
	int j, i, x, c;
	int name, name_mask;
	int v_line;
	uint8 *src;
	int nt_line;
	uint16 *dst;

//	for(j = 0/*(used_sprite_index - 1)*/; j <used_sprite_index/* 0*/; j += 1)
	for(j = (used_sprite_index - 1); j >=0; j -= 1)
	{
		i = used_sprite_list[j];
		p = &sprite_list[i];

		if( (line >= p->top) && (line < p->bottom))
		{
			v_line = (line - p->top) & p->height;
			nt_line = v_line;
			if(p->flags & FLAG_YFLIP) nt_line = (p->height - nt_line);
			name_mask = ((nt_line >> 4) & 3) << 1;
			name = (p->name_left | name_mask);
			v_line &= 0x0F;

			src = &obj_pattern_cache[(name << 8) | ((v_line & 0x0f) << 4)];
//			dst = (uint16 *)&bitmap.data[(line * bitmap.pitch) + (((/*0x20+*/p->xpos) & 0x1ff) * (bitmap.granularity))];
			dst = Line_Cache + (p->xpos /*& 0x1ff*/) + 0x20;
//				(uint16 *)&bitmap.data[(line * bitmap.pitch) + (((/*0x20+*/p->xpos) & 0x1ff) * (bitmap.granularity))];


			for(x = 0; x < 0x10; x += 1)
			{
				c = src[x];
				if(p->flags & FLAG_PRIORITY){	// BG D
					if(c && (((dst[x]&0x0f) == 0) ))//|| (dst[x]&0x0100)))
//					if(c && (((dst[x]&0x0f) == 0 && !(dst[x] & 0x100)) || (dst[x] & 0x100)))
//					if(c && (!(dst[x]&0x10f) || (dst[x]&0x100)))
						dst[x] = 0x100 | ((c) | p->palette);
				}
				else{							// sprite
					if(c)
						dst[x] = 0x100 | ((c) | p->palette);
				}
			}
			if(p->flags & FLAG_CGX)
			{
				name = (p->name_right | name_mask);
				src = &obj_pattern_cache[(name << 8) | ((v_line & 0x0f) << 4)];
				dst += 0x10;

				for(x = 0; x < 0x10; x += 1)
				{
					c = src[x];
					if(p->flags & FLAG_PRIORITY){	// BG D
						if(c && ((dst[x]&0x0f) == 0))
//						if(c && ((dst[x]&0x10f) == 0 || (dst[x] & 0x100)))
							dst[x] = 0x100 | ((c) | p->palette);
					}
					else{
						if(c)
							dst[x] = 0x100 | ((c) | p->palette);
					}
				}
			}
		}
	}
}





void PCE::KailleraClientProc(){
	if(nnnKailleraFlag){
		if(KailleraCounter){
			if(KailleraCounter != NESTER_settings.nes.preferences.KailleraFrameInterval){
				*Pad[0] = nnnKaillerapad[0];
				*Pad[1] = nnnKaillerapad[1];
				*Pad[2] = nnnKaillerapad[2];
				*Pad[3] = nnnKaillerapad[3];
				*Pad[4] = nnnKaillerapad[4];
				KailleraCounter++;
				return;
			}
			KailleraCounter=0;
		}
		unsigned char buff[8];
		int recv;

		memset(buff, 0, sizeof(buff));
		buff[0] = *Pad[0];
		if(nnnKailleraplayer){
			if(-1 == (recv= nnnkailleraModifyPlayValues(&buff,1))){
//				nnnkailleraEndGame();
				nnnKailleraFlag=0;
				return;
			}
		}
		else{
			if(-1 == (recv= nnnkailleraModifyPlayValues(&buff,0))){
//				nnnkailleraEndGame();
				nnnKailleraFlag=0;
				return;
			}
		}
		for(int i=0; i<recv; i++){
			*Pad[i] = nnnKaillerapad[i] = buff[i];
		}
		KailleraCounter++;
	}
	return;
}



#ifdef PCE_FAST_MEM
extern "C" {
void io_page_w(int address, int data);
int io_page_r(int address);
void vdc_w(int address, int data);
void bank_set(int bank, int value);
}

void bank_set(int bank, int value){
	g_PCE->bank_set(bank, value);
}

void vdc_w(int address, int data){
	g_PCE->vdc_w(address, data);
}

void io_page_w(int address, int data){
#ifdef _DEBUG
//	LOG("io_page_w" << HEX(address,8) <<"  data " << HEX(data,8) <<endl);
#endif
	g_PCE->io_page_w(address, data);
}

int io_page_r(int address){
	return g_PCE->io_page_r(address);
}


#else 
extern "C" {
	void cpu_writeport16(int port, int data);
	void cpu_writemem21(int address, int data);
	int cpu_readmem21(int address);
}


void cpu_writeport16(int port, int data){
	g_PCE->cpu_writeport16(port, data);
}

void cpu_writemem21(int address, int data){
	g_PCE->cpu_writemem21(address, data);
}

int cpu_readmem21(int address){
	return g_PCE->cpu_readmem21(address);
}
#endif

#endif // NES_ONLY
/*
*/